#if !defined HAVE_PERM_TROTTER_H__
#define      HAVE_PERM_TROTTER_H__

#include "aux0/swap.h"
#include "comb/comb-print.h"
#include "fxttypes.h"


#define TROTTER_OPT  // much faster computations
// XOPT takes effect only if OPT is defined:
#define TROTTER_XOPT  // slightly faster computations

class perm_trotter
// Gray code for permutations
// Trotter/Johnson algorithm.
// Smalles element moves most often.
{
public:
    ulong n_;     // number of elements to permute
    ulong *xi_;   // inverse permutation
    ulong *x_;    // permutation of {0, 1, ..., n-1}
    ulong *d_;    // auxiliary: directions
    ulong sw1_, sw2_; // indices of elements swapped most recently

public:
    perm_trotter(ulong n)
    {
        n_ = n;
        xi_ = new ulong[n_];
        x_ = new ulong[n_+2];
        d_ = new ulong[n_];
        ulong sen = 0;  // sentinel value minimal
        x_[0] = x_[n_+1] = sen;
        ++x_;
        first();
    }

    ~perm_trotter()
    {
        --x_;
        delete [] xi_;
        delete [] d_;
        delete [] x_;
    }

    void first()
    {
        for (ulong i=0; i<n_; i++)  xi_[i] = i;
        for (ulong i=0; i<n_; i++)  x_[i] = i;
        for (ulong i=0; i<n_; i++)  d_[i] = 1;
        sw1_ = n_ - 1;  sw2_ = n_ - 2;  // relative to last permutation
    }

    void last()
    {
        for (ulong i=0; i<n_; i++)  xi_[i] = i;
        for (ulong i=0; i<n_; i++)  x_[i] = i;
        for (ulong i=0; i<n_; i++)  d_[i] = -1UL;
        sw1_ = n_ - 1;  sw2_ = n_ - 2;  // relative to first permutation
        d_[sw1_] = +1;
        d_[sw2_] = +1;
        swap2(x_[sw1_], x_[sw2_]);
        swap2(xi_[sw1_], xi_[sw2_]);
    }

    bool next()
    {
#ifdef TROTTER_OPT
        {  // most frequent case: e1 == 0
            ulong i1 = xi_[0];   // position of element e1
            ulong d = d_[0];     // direction to move e1
            ulong i2 = i1 + d;   // position to swap with
            ulong e2 = x_[i2];   // element to swap with

            if ( 0 < e2 )  // can we swap?
            {
                xi_[0] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = 0;
                sw1_ = i1;  sw2_ = i2;
                return true;
            }
        }

#ifdef TROTTER_XOPT
        {  // second most frequent case: e1 == 1
            ulong i1 = xi_[1];   // position of element e1
            ulong d = d_[1];     // direction to move e1
            ulong i2 = i1 + d;   // position to swap with
            ulong e2 = x_[i2];   // element to swap with

            if ( 1 < e2 )  // can we swap?
            {
                xi_[1] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = 1;
                d_[0] = -d_[0];  // negate
                sw1_ = i1;  sw2_ = i2;
                return true;
            }
        }
#endif // TROTTER_XOPT
#endif // TROTTER_OPT

#ifdef TROTTER_OPT
#ifdef TROTTER_XOPT
        for (ulong e1=2; e1<n_; ++e1)  // note: start at e1=2
#else
        for (ulong e1=1; e1<n_; ++e1)  // note: start at e1=1
#endif  // TROTTER_XOPT
#else   // TROTTER_OPT
        for (ulong e1=0; e1<n_; ++e1)
#endif // TROTTER_OPT
        {
            // e1 is the element we try to move
            ulong i1 = xi_[e1];   // position of element e1
            ulong d = d_[e1];     // direction to move e1
            ulong i2 = i1 + d;    // position to swap with
            ulong e2 = x_[i2];    // element to swap with

            if ( e1 < e2 )  // can we swap?
            {
                xi_[e1] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = e1;
                sw1_ = i1;  sw2_ = i2;
                while ( e1-- )  d_[e1] = -d_[e1];
                return true;
            }
        }

        first();
        return false;
    }


    bool prev()
    {
#ifdef TROTTER_OPT
        {  // most frequent case: e1 == 0
            ulong i1 = xi_[0];   // position of element e1
            ulong d = -d_[0];    // direction to move e1 (NOTE: negated)
            ulong i2 = i1 + d;   // position to swap with
            ulong e2 = x_[i2];   // element to swap with

            if ( 0 < e2 )  // can we swap?
            {
                xi_[0] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = 0;
                sw1_ = i1;  sw2_ = i2;
                return true;
            }
        }
#ifdef TROTTER_XOPT
        {  // second most frequent case: e1 == 1
            ulong i1 = xi_[1];   // position of element e1
            ulong d = -d_[1];    // direction to move e1 (NOTE: negated)
            ulong i2 = i1 + d;   // position to swap with
            ulong e2 = x_[i2];   // element to swap with

            if ( 1 < e2 )  // can we swap?
            {
                xi_[1] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = 1;
                d_[0] = -d_[0];  // negate
                sw1_ = i1;  sw2_ = i2;
                return true;
            }
        }
#endif // TROTTER_XOPT
#endif // TROTTER_OPT

#ifdef TROTTER_OPT
#ifdef TROTTER_XOPT
        for (ulong e1=2; e1<n_; ++e1)  // note: start at e1=2
#else
        for (ulong e1=1; e1<n_; ++e1)  // note: start at e1=1
#endif  // TROTTER_XOPT
#else   // TROTTER_OPT
        for (ulong e1=0; e1<n_; ++e1)
#endif // TROTTER_OPT
        {
            // e1 is the element we try to move
            ulong i1 = xi_[e1];   // position of element e1
            ulong d = -d_[e1];    // direction to move e1 (NOTE: negated)
            ulong i2 = i1 + d;    // position to swap with
            ulong e2 = x_[i2];    // element to swap with

            if ( e1 < e2 )  // can we swap?
            {
                xi_[e1] = i2;
                xi_[e2] = i1;
                x_[i1] = e2;
                x_[i2] = e1;
                sw1_ = i1;  sw2_ = i2;
                while ( e1-- )  d_[e1] = -d_[e1];
                return true;
            }
        }

        last();
        return false;
    }

    const ulong * data()  const  { return x_; }
    const ulong * invdata()  const  { return xi_; }
    void get_swap(ulong &s1, ulong &s2)  const  { s1=sw1_; s2=sw2_; }

    void print(const char *bla, bool dfz=false)  const
    // If dfz is true then Dots are printed For Zeros.
    { print_perm(bla, data(), n_, dfz); }

    void print_inv(const char *bla, bool dfz=false)  const
    // If dfz is true then Dots are printed For Zeros.
    { print_perm(bla, invdata(), n_, dfz); }
};
// -------------------------

#undef TROTTER_OPT
#undef TROTTER_XOPT


#endif  // !defined HAVE_PERM_TROTTER_H__
